/*++

Module Name:

    engine.c


Core engine functions.

Overview:

	This module handles the actual execution of scripts. 
	

--*/

// #define DBG 1

#include <WDM.H>
#include "filter.h"


void LoadConfiguration( struct DEVICE_EXTENSION* DevExt, PCONFIG NewConfig)
{
	PCONFIG Current = NewConfig;
	while (Current != NULL)
	{
		DevExt->Buttons[(Current->Button)/2][(Current->Button)%2] = Current->NewCommand;
		Current = Current->Next;
	}
}


VOID ExecuteScript( IN PKDPC Dpc, IN PVOID DeferredContext, IN PVOID SystemArgument1, IN PVOID SystemArgument2)
{
struct DEVICE_EXTENSION * DevExt = (struct DEVICE_EXTENSION *)DeferredContext;
    LARGE_INTEGER SysTime;
    QENTRY CurrentQ; PSCRIPT Current;
    unsigned char Op;  unsigned short Offset;
//    SCRIPT Temp, Next;

	// Dequeue items.
	KeQuerySystemTime(&SysTime);
	SysTime = RtlLargeIntegerAdd(SysTime,RtlConvertLongToLargeInteger(20)); // Offset because of Windows' timer inaccuracy.
	
	QueuePeek(&DevExt->ActionQueue, &CurrentQ);
	while ((CurrentQ.Head != NULL) && (RtlLargeIntegerGreaterThan(SysTime,CurrentQ.Time)))
	{
		QueuePop(&DevExt->ActionQueue, &CurrentQ);
		Current = CurrentQ.Current;
		
		if (Current != NULL)
		{
			DbgPrn(("Executing item:   Op: %d   Offset: %d   Time: %x %x\n",Current->Op,Current->Offset,CurrentQ.Time.HighPart,CurrentQ.Time.LowPart));		
			Op = Current->Op;
			Offset = Current->Offset;
			
			if (Op & 0x04)  // 2.x opcodes.
			{
				switch (Op&0x03)
				{
					case 0: // Windows hat switch command.
						Offset=(Offset>8)?0:Offset;
						DevExt->KeyboardStatus[24]=(DevExt->KeyboardStatus[24]&0x0F)+(Offset<<4);
						DevExt->NewData = DevExt->NewData | 1; // New stick data
						break;
					case 1: // Load new configuration.
						Offset = Offset&0xFF;
						LoadConfiguration(DevExt,DevExt->Configurations[Offset]);
						break;						
					case 2: // If frobstatus set, jump to command x.
					case 3: // If frobstatus unset, jump to command x.
					default:
						// Do nothing.
						;
				}
				
			}
			else  // Regular command.
			{
				if (Op & 0x02)
				{
					if (Offset > 32) Offset = 32;
					Offset+=(21*8)-1;
					DevExt->NewData = DevExt->NewData | 1; // New stick data
				}
				else
					DevExt->NewData = DevExt->NewData | 2; // New keybd data
				
				if (Op & 0x01)
				{
					// Release.
					DevExt->KeyboardStatus[(Offset)/8]=DevExt->KeyboardStatus[(Offset)/8] & ~(1 << (Offset%8));
				}
				else
				{
					// Press.
					DevExt->KeyboardStatus[(Offset)/8]=DevExt->KeyboardStatus[(Offset)/8] | (1 << (Offset%8));
				}
			}
			
			if (Current->Next != NULL)
			{
				DbgPrn(("Queuing next item in comand.\n"));
				CurrentQ.Current = Current->Next;
				CurrentQ.Time = RtlLargeIntegerAdd(CurrentQ.Time, RtlConvertLongToLargeInteger(CurrentQ.Current->Delay));
				AddScript(&DevExt->ActionQueue, CurrentQ);
			}
		}
		QueuePeek(&DevExt->ActionQueue, &CurrentQ);
	}
	
	
	// Reset timer.
	if (CurrentQ.Head)
		KeSetTimer( &DevExt->Timer, CurrentQ.Time, &DevExt->TimerDPC );              
		

  	if (DevExt->NewData)
  	{
	  		KeInsertQueueDpc(&DevExt->ReturnDPC,NULL,NULL);
  	}

}
